import { dataSource } from "../../../data-source";
import { equipment } from "../../../entity/configuration/equipment";
import DatabaseActionException from "../../../exceptions/databaseActionException";
import { CreateEquipmentCommand, DeleteEquipmentCommand, UpdateEquipmentCommand } from "./equipmentCommand";

export default abstract class EquipmentCommandHandler {
  /**
   * @description create equipment
   * @param {CreateEquipmentCommand} createEquipment
   * @returns {Promise<string>}
   */
  static async create(createEquipment: CreateEquipmentCommand): Promise<string> {
    return await dataSource
      .createQueryBuilder()
      .insert()
      .into(equipment)
      .values({
        code: createEquipment.code,
        type: createEquipment.type,
        name: createEquipment.name,
        commissioned: createEquipment.commissioned,
        decommissioned: createEquipment.decommissioned,
      })
      .execute()
      .then((result) => {
        return result.identifiers[0].id;
      })
      .catch((err) => {
        throw new DatabaseActionException("CREATE", "equipment", err);
      });
  }

  /**
   * @description update equipment
   * @param {UpdateEquipmentCommand} updateEquipment
   * @returns {Promise<void>}
   */
  static async update(updateEquipment: UpdateEquipmentCommand): Promise<void> {
    return await dataSource
      .createQueryBuilder()
      .update(equipment)
      .set({
        code: updateEquipment.code,
        type: updateEquipment.type,
        name: updateEquipment.name,
        commissioned: updateEquipment.commissioned,
        decommissioned: updateEquipment.decommissioned,
      })
      .where("id = :id", { id: updateEquipment.id })
      .execute()
      .then(() => {})
      .catch((err) => {
        throw new DatabaseActionException("UPDATE", "equipment", err);
      });
  }

  /**
   * @description decommission equipment
   * @param {string} equipmentId
   * @returns {Promise<void>}
   */
  static async decommission(equipmentId: string): Promise<void> {
    return await dataSource
      .createQueryBuilder()
      .update(equipment)
      .set({
        decommissioned: new Date(),
      })
      .where("id = :id", { id: equipmentId })
      .execute()
      .then(() => {})
      .catch((err) => {
        throw new DatabaseActionException("UPDATE", "equipment", err);
      });
  }

  /**
   * @description delete equipment
   * @param {DeleteEquipmentCommand} deleteEquipment
   * @returns {Promise<void>}
   */
  static async delete(deleteEquipment: DeleteEquipmentCommand): Promise<void> {
    return await dataSource
      .createQueryBuilder()
      .delete()
      .from(equipment)
      .where("id = :id", { id: deleteEquipment.id })
      .execute()
      .then(() => {})
      .catch((err) => {
        throw new DatabaseActionException("DELETE", "equipment", err);
      });
  }
}
