import { dataSource } from "../../../data-source";
import { force } from "../../../entity/configuration/force";
import DatabaseActionException from "../../../exceptions/databaseActionException";
import { CreateForceCommand, DeleteForceCommand, UpdateForceCommand } from "./forceCommand";

export default abstract class ForceCommandHandler {
  /**
   * @description create force
   * @param {CreateForceCommand} createForce
   * @returns {Promise<string>}
   */
  static async create(createForce: CreateForceCommand): Promise<string> {
    return await dataSource
      .createQueryBuilder()
      .insert()
      .into(force)
      .values({
        internalId: createForce.internalId,
        firstname: createForce.firstname,
        lastname: createForce.lastname,
        nameaffix: createForce.nameaffix,
        commissioned: createForce.commissioned,
        decommissioned: createForce.decommissioned,
      })
      .execute()
      .then((result) => {
        return result.identifiers[0].id;
      })
      .catch((err) => {
        throw new DatabaseActionException("CREATE", "force", err);
      });
  }

  /**
   * @description update force
   * @param {UpdateForceCommand} updateForce
   * @returns {Promise<void>}
   */
  static async update(updateForce: UpdateForceCommand): Promise<void> {
    return await dataSource
      .createQueryBuilder()
      .update(force)
      .set({
        internalId: updateForce.internalId,
        firstname: updateForce.firstname,
        lastname: updateForce.lastname,
        nameaffix: updateForce.nameaffix,
        commissioned: updateForce.commissioned,
        decommissioned: updateForce.decommissioned,
      })
      .where("id = :id", { id: updateForce.id })
      .execute()
      .then(() => {})
      .catch((err) => {
        throw new DatabaseActionException("UPDATE", "force", err);
      });
  }

  /**
   * @description decommission force
   * @param {string} forceId
   * @returns {Promise<void>}
   */
  static async decommission(forceId: string): Promise<void> {
    return await dataSource
      .createQueryBuilder()
      .update(force)
      .set({
        decommissioned: new Date(),
      })
      .where("id = :id", { id: forceId })
      .execute()
      .then(() => {})
      .catch((err) => {
        throw new DatabaseActionException("UPDATE", "force", err);
      });
  }

  /**
   * @description delete force
   * @param {DeleteForceCommand} deleteForce
   * @returns {Promise<void>}
   */
  static async delete(deleteForce: DeleteForceCommand): Promise<void> {
    return await dataSource
      .createQueryBuilder()
      .delete()
      .from(force)
      .where("id = :id", { id: deleteForce.id })
      .execute()
      .then(() => {})
      .catch((err) => {
        throw new DatabaseActionException("DELETE", "force", err);
      });
  }
}
