import { Request, Response } from "express";
import ForceService from "../../../service/configuration/forceService";
import ForceFactory from "../../../factory/admin/configuration/force";
import {
  CreateForceCommand,
  DeleteForceCommand,
  UpdateForceCommand,
} from "../../../command/configuration/force/forceCommand";
import ForceCommandHandler from "../../../command/configuration/force/forceCommandHandler";
import MissionService from "../../../service/operation/missionService";

/**
 * @description get all forces
 * @param req {Request} Express req object
 * @param res {Response} Express res object
 * @returns {Promise<*>}
 */
export async function getAllForces(req: Request, res: Response): Promise<any> {
  let offset = parseInt((req.query.offset as string) ?? "0");
  let count = parseInt((req.query.count as string) ?? "25");
  let search = (req.query.search as string) ?? "";
  let noLimit = req.query.noLimit === "true";
  let ids = ((req.query.ids ?? "") as string).split(",").filter((i) => i);

  let [forces, total] = await ForceService.getAll({ offset, count, search, noLimit, ids });

  res.json({
    forces: ForceFactory.mapToBase(forces),
    total: total,
    offset: offset,
    count: count,
  });
}

/**
 * @description get forces by Ids
 * @param req {Request} Express req object
 * @param res {Response} Express res object
 * @returns {Promise<*>}
 */
export async function getForcesByIds(req: Request, res: Response): Promise<any> {
  let ids = req.body.ids as Array<string>;

  let [forces, total] = await ForceService.getAll({ noLimit: true, ids });

  res.json({
    forces: ForceFactory.mapToBase(forces),
    total: total,
    offset: 0,
    count: total,
  });
}

/**
 * @description get force by id
 * @param req {Request} Express req object
 * @param res {Response} Express res object
 * @returns {Promise<*>}
 */
export async function getForceById(req: Request, res: Response): Promise<any> {
  const forceId = req.params.id;
  let force = await ForceService.getById(forceId);

  res.json(ForceFactory.mapToSingle(force));
}

/**
 * @description get available forces by missionId
 * @param req {Request} Express req object
 * @param res {Response} Express res object
 * @returns {Promise<*>}
 */
export async function getAvailabeForcesByMission(req: Request, res: Response): Promise<any> {
  const missionId = req.params.missionId;
  let mission = await MissionService.getById(missionId);
  let forces = await ForceService.getAvailable(mission.mission_start ?? new Date());

  res.json(ForceFactory.mapToBase(forces));
}

/**
 * @description create force
 * @param req {Request} Express req object
 * @param res {Response} Express res object
 * @returns {Promise<*>}
 */
export async function createForce(req: Request, res: Response): Promise<any> {
  const internalId = req.body.internalId || null;
  const firstname = req.body.firstname;
  const lastname = req.body.lastname;
  const nameaffix = req.body.nameaffix;
  const commissioned = req.body.commissioned;
  const decommissioned = req.body.decommissioned || null;

  let createForce: CreateForceCommand = {
    firstname,
    lastname,
    nameaffix,
    internalId,
    commissioned,
    decommissioned,
  };
  let forceId = await ForceCommandHandler.create(createForce);

  res.status(200).send(forceId);
}

/**
 * @description update force by id
 * @param req {Request} Express req object
 * @param res {Response} Express res object
 * @returns {Promise<*>}
 */
export async function updateForceById(req: Request, res: Response): Promise<any> {
  const forceId = req.params.id;
  const internalId = req.body.internalId || null;
  const firstname = req.body.firstname;
  const lastname = req.body.lastname;
  const nameaffix = req.body.nameaffix;
  const commissioned = req.body.commissioned;
  const decommissioned = req.body.decommissioned || null;

  let updateForce: UpdateForceCommand = {
    id: forceId,
    firstname,
    lastname,
    nameaffix,
    internalId,
    commissioned,
    decommissioned,
  };
  await ForceCommandHandler.update(updateForce);

  res.sendStatus(204);
}

/**
 * @description decommission force by id
 * @param req {Request} Express req object
 * @param res {Response} Express res object
 * @returns {Promise<*>}
 */
export async function decommissionForceById(req: Request, res: Response): Promise<any> {
  const forceId = req.params.id;

  await ForceCommandHandler.decommission(forceId);

  res.sendStatus(204);
}

/**
 * @description delete force by id
 * @param req {Request} Express req object
 * @param res {Response} Express res object
 * @returns {Promise<*>}
 */
export async function deleteForceById(req: Request, res: Response): Promise<any> {
  const forceId = req.params.id;

  let deleteForce: DeleteForceCommand = {
    id: forceId,
  };
  await ForceCommandHandler.delete(deleteForce);

  res.sendStatus(204);
}
