import { dataSource } from "../data-source";
import { FileSystemHelper } from "./fileSystemHelper";
import { EntityManager } from "typeorm";
import uniqBy from "lodash.uniqby";
import InternalException from "../exceptions/internalException";
import UserService from "../service/management/userService";
import { BACKUP_COPIES, BACKUP_INTERVAL } from "../env.defaults";
import DatabaseActionException from "../exceptions/databaseActionException";

export type BackupSection = "force" | "user";

export type BackupSectionRefered = {
  [key in BackupSection]?: Array<string>;
};

export type BackupFileContent = { [key in BackupSection]?: BackupFileContentSection } & {
  backup_file_details: { collectIds: boolean; createdAt: Date; version: 1 };
};
export type BackupFileContentSection = Array<any> | { [key: string]: Array<any> };

export default abstract class BackupHelper {
  // ! Order matters because of foreign keys
  private static readonly backupSection: Array<{ type: BackupSection; orderOnInsert: number; orderOnClear: number }> = [
    { type: "force", orderOnInsert: 2, orderOnClear: 2 },
    { type: "user", orderOnInsert: 1, orderOnClear: 1 },
  ];

  private static readonly backupSectionRefered: BackupSectionRefered = {
    force: ["force"],
    user: ["user", "user_permission", "role", "role_permission", "invite"],
  };

  private static transactionManager: EntityManager;

  static async createBackup({
    filename,
    path = "/backup",
    collectIds = true,
  }: {
    filename?: string;
    path?: string;
    collectIds?: boolean;
  }): Promise<void> {
    if (!filename) {
      filename = new Date().toISOString().split("T")[0];
    }

    let json: BackupFileContent = { backup_file_details: { collectIds, createdAt: new Date(), version: 1 } };
    for (const section of this.backupSection) {
      json[section.type] = await this.getSectionData(section.type, collectIds);
    }

    FileSystemHelper.writeFile(path, filename + ".json", JSON.stringify(json, null, 2));

    let files = FileSystemHelper.getFilesInDirectory("backup", ".json");
    let sorted = files.sort((a, b) => new Date(b.split(".")[0]).getTime() - new Date(a.split(".")[0]).getTime());

    const filesToDelete = sorted.slice(BACKUP_COPIES);
    for (const file of filesToDelete) {
      FileSystemHelper.deleteFile("backup", file);
    }
  }

  static async createBackupOnInterval() {
    let files = FileSystemHelper.getFilesInDirectory("backup", ".json");
    let newestFile = files.sort((a, b) => new Date(b.split(".")[0]).getTime() - new Date(a.split(".")[0]).getTime())[0];

    let lastBackup = new Date(newestFile.split(".")[0]);
    let diffInMs = new Date().getTime() - lastBackup.getTime();
    let diffInDays = diffInMs / (1000 * 60 * 60 * 24);

    if (diffInDays >= BACKUP_INTERVAL) {
      await this.createBackup({});
    }
  }

  static async loadBackup({
    filename,
    path = "/backup",
    include = [],
    partial = false,
    overwrite = false,
  }: {
    filename: string;
    path?: string;
    partial?: boolean;
    include?: Array<BackupSection>;
    overwrite?: boolean;
  }): Promise<void> {
    this.transactionManager = undefined;

    let file = FileSystemHelper.readFile(`${path}/${filename}`);
    let backup: BackupFileContent = JSON.parse(file);

    if ((partial && include.length == 0) || (!partial && include.length != 0)) {
      throw new InternalException("partial and include have to be set correctly for restoring backup.");
    }

    await dataSource.manager
      .transaction(async (transaction) => {
        this.transactionManager = transaction;

        const sections = this.backupSection
          .filter((bs) => (partial ? include.includes(bs.type) : true))
          .sort((a, b) => a.orderOnClear - b.orderOnClear);
        if (!overwrite) {
          for (const section of sections.filter((s) => Object.keys(backup).includes(s.type))) {
            let refered = this.backupSectionRefered[section.type];
            for (const ref of refered) {
              await this.transactionManager.getRepository(ref).delete({});
            }
          }
        }

        for (const section of sections
          .filter((s) => Object.keys(backup).includes(s.type))
          .sort((a, b) => a.orderOnInsert - b.orderOnInsert)) {
          await this.setSectionData(section.type, backup[section.type], backup.backup_file_details.collectIds ?? false);
        }

        this.transactionManager = undefined;
      })
      .catch((err) => {
        console.log(err);
        this.transactionManager = undefined;
        throw new DatabaseActionException("BACKUP RESTORE", include.join(", ") || "FULL", err);
      });
  }

  public static async autoRestoreBackup() {
    let count = await UserService.count();
    if (count == 0) {
      let files = FileSystemHelper.getFilesInDirectory("/backup", ".json");
      let newestFile = files.sort(
        (a, b) => new Date(b.split(".")[0]).getTime() - new Date(a.split(".")[0]).getTime()
      )[0];

      if (newestFile) {
        console.log(`${new Date().toISOString()}: auto-restoring ${newestFile}`);
        await this.loadBackup({ filename: newestFile });
        console.log(`${new Date().toISOString()}: finished auto-restore`);
      } else {
        console.log(`${new Date().toISOString()}: skip auto-restore as no backup was found`);
      }
    } else {
      console.log(`${new Date().toISOString()}: skip auto-restore as users exist`);
    }
  }

  private static async getSectionData(
    section: BackupSection,
    collectIds: boolean
  ): Promise<Array<any> | { [key: string]: any }> {
    switch (section) {
      case "force":
        return await this.getForceData(collectIds);
      case "user":
        return await this.getUser(collectIds);
      default:
        return [];
    }
  }

  private static async getForceData(collectIds: boolean): Promise<Array<any>> {
    return await dataSource
      .getRepository("force")
      .createQueryBuilder("force")
      .select([...(collectIds ? ["force.id"] : []), "force.firstname", "force.lastname", "force.nameaffix"])
      .getMany();
  }
  private static async getUser(collectIds: boolean): Promise<{ [key: string]: Array<any> }> {
    return {
      user: await dataSource
        .getRepository("user")
        .createQueryBuilder("user")
        .leftJoin("user.roles", "roles")
        .leftJoin("roles.permissions", "role_permissions")
        .leftJoin("user.permissions", "permissions")
        .select([
          ...(collectIds ? ["user.id"] : []),
          "user.mail",
          "user.username",
          "user.firstname",
          "user.lastname",
          "user.secret",
          "user.isOwner",
        ])
        .addSelect(["permissions.permission"])
        .addSelect(["roles.role"])
        .addSelect(["role_permissions.permission"])
        .getMany(),
      role: await dataSource
        .getRepository("role")
        .createQueryBuilder("role")
        .leftJoin("role.permissions", "permissions")
        .addSelect(["role.role"])
        .addSelect(["permissions.permission"])
        .getMany(),
      invite: await dataSource.getRepository("invite").find(),
    };
  }

  private static async setSectionData(
    section: BackupSection,
    data: BackupFileContentSection,
    collectedIds: boolean
  ): Promise<void> {
    if (section == "force" && Array.isArray(data)) await this.setForceData(data);
    if (section == "user" && !Array.isArray(data)) await this.setUser(data);
  }

  private static async setForceData(data: Array<any>): Promise<void> {
    await this.transactionManager.getRepository("force").save(data);
  }
  private static async setUser(data: { [key: string]: Array<any> }): Promise<void> {
    let usedRoles = (data?.["user"] ?? [])
      .map((d) => d.roles)
      .flat()
      .map((d) => ({ ...d, id: undefined }));

    await this.transactionManager
      .createQueryBuilder()
      .insert()
      .into("role")
      .values(uniqBy([...(data?.["role"] ?? []), ...usedRoles], "role"))
      .orIgnore()
      .execute();

    let roles = await this.transactionManager.getRepository("role").find();
    let dataWithMappedIds = (data?.["user"] ?? []).map((u) => ({
      ...u,
      roles: u.roles.map((r: any) => ({
        ...r,
        id: roles.find((role) => role.role == r.role)?.id ?? undefined,
      })),
    }));
    await this.transactionManager.getRepository("user").save(dataWithMappedIds);
    await this.transactionManager
      .createQueryBuilder()
      .insert()
      .into("invite")
      .values(data["invite"])
      .orIgnore()
      .execute();
  }
}
