import * as Y from "yjs";
import { MissionMap } from "../storage/missionMap";
import MissionService from "../service/operation/missionService";
import { SyncMissionCommand } from "../command/operation/mission/missionCommand";
import MissionCommandHandler from "../command/operation/mission/missionCommandHandler";
import MissionVehicleCommandHandler from "../command/operation/mission/missionVehicleCommandHandler";
import MissionEquipmentCommandHandler from "../command/operation/mission/missionEquipmentCommandHandler";
import MissionContactCommandHandler from "../command/operation/mission/missionContactCommandHandler";
import MissionPresenceCommandHandler from "../command/operation/mission/missionPresenceCommandHandler";
import differenceWith from "lodash.differencewith";
import isEqual from "lodash.isequal";

export default abstract class MissionDocHelper {
  public static async populateDoc(missionId: string) {
    const mission = await MissionService.getFullById(missionId);

    const doc = new Y.Doc();
    doc.getMap("form").set("title", mission.title);
    doc.getMap("form").set("command", mission.commandId ?? "");
    doc.getMap("form").set("secretary", mission.secretaryId ?? "");
    doc.getMap("form").set("start", mission.mission_start?.toISOString() ?? "");
    doc.getMap("form").set("end", mission.mission_end?.toISOString() ?? "");
    doc.getMap("form").set("mission_short", mission.keyword);
    doc.getMap("form").set("location", mission.location);
    doc.getMap("form").set("others", mission.others);
    doc.getMap("form").set("rescued", mission.rescued.toString());
    doc.getMap("form").set("recovered", mission.recovered.toString());
    doc.getText("editor").applyDelta(JSON.parse(mission.description));

    for (const vehicle of mission.vehicles) {
      doc.getMap<Y.Map<number | string>>("vehicle").set(vehicle.vehicleId, new Y.Map<string | number>());
      for (const key of ["driverId", "leaderId", "mileage_start", "mileage_end"] as (keyof typeof vehicle)[]) {
        doc
          .getMap<Y.Map<number | string>>("vehicle")
          .get(vehicle.vehicleId)
          .set<string | number>(key.replace("Id", ""), (vehicle[key] ?? "").toString());
      }
    }
    for (const equipment of mission.equipments) {
      doc.getMap<Y.Map<number | string>>("equipment").set(equipment.equipmentId, new Y.Map<string | number>());
      for (const key of ["note"] as (keyof typeof equipment)[]) {
        doc
          .getMap<Y.Map<number | string>>("equipment")
          .get(equipment.equipmentId)
          .set<string | number>(key, (equipment[key] ?? "").toString());
      }
    }
    for (const contact of mission.contacts) {
      doc.getMap<Y.Map<number | string>>("contact").set(contact.contactId, new Y.Map<string | number>());
      for (const key of ["firstname", "lastname", "phone", "address", "note"] as (keyof typeof contact)[]) {
        doc
          .getMap<Y.Map<number | string>>("contact")
          .get(contact.contactId)
          .set<string | number>(key, (contact[key] ?? "").toString());
      }
    }
    for (const force of mission.presence) {
      doc.getMap<boolean>("presence").set(force.forceId, true);
    }

    MissionMap.write(missionId, {
      missionId,
      doc,
      editors: new Map(),
      editorStates: new Map(),
      timestamp: mission.last_update,
    });
    console.log(`created local doc ${missionId}`);
  }

  public static docAsUpdate(missionId: string): Uint8Array {
    const mission = MissionMap.read(missionId);
    return Y.encodeStateAsUpdate(mission.doc);
  }

  public static docAsStateVector(missionId: string): Uint8Array {
    const mission = MissionMap.read(missionId);
    return Y.encodeStateVector(mission.doc);
  }

  public static docTimestamp(missionId: string): number {
    const mission = MissionMap.read(missionId);
    return mission.timestamp;
  }

  public static getAwarenessState(missionId: string) {
    const mission = MissionMap.read(missionId);
    return Array.from(mission.editorStates.entries());
  }

  public static getAwarenessEditors(missionId: string) {
    const mission = MissionMap.read(missionId);
    return Array.from(mission.editors.entries());
  }

  public static async saveDoc(missionId: string, update: Uint8Array, timestamp: number) {
    const oldMission = MissionMap.read(missionId);
    const oldVehicles = Array.from(oldMission.doc.getMap("presence").keys());
    const oldEquipments = Array.from(oldMission.doc.getMap("presence").keys());
    const oldContact = Array.from(oldMission.doc.getMap("presence").keys());
    const oldPresence = Array.from(oldMission.doc.getMap("presence").keys());

    const oldMissionVehicles = oldMission.doc.getMap<Y.Map<string | number>>("vehicle").toJSON();
    const oldMissionEquipment = oldMission.doc.getMap<Y.Map<string | number>>("equipment").toJSON();
    const oldMissionContact = oldMission.doc.getMap<Y.Map<string | number>>("contact").toJSON();
    const oldMissionPresence = oldMission.doc.getMap<Y.Map<string | number>>("presence").toJSON();

    const oldMissionVehiclesArray = Object.keys(oldMissionVehicles).map((mv) => ({
      missionId: missionId,
      vehicleId: mv,
      driverId: oldMissionVehicles[mv]?.driver || null,
      leaderId: oldMissionVehicles[mv]?.leader || null,
      mileage_start: parseInt(oldMissionVehicles[mv]?.mileage_start) || null,
      mileage_end: parseInt(oldMissionVehicles[mv]?.mileage_end) || null,
    }));
    const oldMissionEquipmentArray = Object.keys(oldMissionEquipment).map((me) => ({
      missionId: missionId,
      equipmentId: me,
      note: oldMissionEquipment[me]?.note ?? "",
    }));
    const oldMissionContactArray = Object.keys(oldMissionContact).map((mc) => ({
      missionId: missionId,
      contactId: mc,
      firstname: oldMissionContact[mc]?.firstname ?? "",
      lastname: oldMissionContact[mc]?.lastname ?? "",
      phone: oldMissionContact[mc]?.phone ?? "",
      address: oldMissionContact[mc]?.address ?? "",
      note: oldMissionContact[mc]?.note ?? "",
    }));
    const oldMissionPresenceArray = Object.keys(oldMissionPresence).map((mc) => ({
      missionId: missionId,
      forceId: mc,
    }));

    MissionMap.updateState(missionId, update);

    const mission = MissionMap.read(missionId);
    const vehicles = Array.from(mission.doc.getMap("presence").keys());
    const equipments = Array.from(mission.doc.getMap("presence").keys());
    const contact = Array.from(mission.doc.getMap("presence").keys());
    const presence = Array.from(mission.doc.getMap("presence").keys());

    const removedVehicles = oldVehicles.filter((item) => !vehicles.includes(item));
    const removedEquipments = oldEquipments.filter((item) => !equipments.includes(item));
    const removedContacts = oldContact.filter((item) => !contact.includes(item));
    const removedPresence = oldPresence.filter((item) => !presence.includes(item));

    const missionDetails: SyncMissionCommand = {
      id: missionId,
      title: mission.doc.getMap("form").toJSON().title,
      ...(mission.doc.getMap("form").toJSON().command != ""
        ? { commandId: mission.doc.getMap("form").toJSON().command }
        : { commandId: null }),
      ...(mission.doc.getMap("form").toJSON().secretary != ""
        ? { secretaryId: mission.doc.getMap("form").toJSON().secretary }
        : { secretaryId: null }),
      ...(mission.doc.getMap("form").get("start") != ""
        ? { mission_start: new Date(mission.doc.getMap("form").get("start") as string) }
        : { mission_start: null }),
      ...(mission.doc.getMap("form").get("end") != ""
        ? { mission_end: new Date(mission.doc.getMap("form").get("end") as string) }
        : { mission_end: null }),
      keyword: mission.doc.getMap("form").toJSON().mission_short ?? "",
      location: mission.doc.getMap("form").toJSON().location ?? "",
      others: mission.doc.getMap("form").toJSON().others ?? "",
      rescued: parseInt(mission.doc.getMap("form").toJSON().rescued) || 0,
      recovered: parseInt(mission.doc.getMap("form").toJSON().recovered) || 0,
      description: JSON.stringify(mission.doc.getText("editor").toDelta()),
      last_update: timestamp,
    };
    const missionVehicles = mission.doc.getMap<Y.Map<string | number>>("vehicle").toJSON();
    const missionEquipment = mission.doc.getMap<Y.Map<string | number>>("equipment").toJSON();
    const missionContact = mission.doc.getMap<Y.Map<string | number>>("contact").toJSON();
    const missionPresence = mission.doc.getMap<Y.Map<string | number>>("presence").toJSON();

    const missionVehiclesArray = Object.keys(missionVehicles).map((mv) => ({
      missionId: missionId,
      vehicleId: mv,
      driverId: missionVehicles[mv]?.driver || null,
      leaderId: missionVehicles[mv]?.leader || null,
      mileage_start: parseInt(missionVehicles[mv]?.mileage_start) || null,
      mileage_end: parseInt(missionVehicles[mv]?.mileage_end) || null,
    }));
    const missionEquipmentArray = Object.keys(missionEquipment).map((me) => ({
      missionId: missionId,
      equipmentId: me,
      note: missionEquipment[me]?.note ?? "",
    }));
    const missionContactArray = Object.keys(missionContact).map((mc) => ({
      missionId: missionId,
      contactId: mc,
      firstname: missionContact[mc]?.firstname ?? "",
      lastname: missionContact[mc]?.lastname ?? "",
      phone: missionContact[mc]?.phone ?? "",
      address: missionContact[mc]?.address ?? "",
      note: missionContact[mc]?.note ?? "",
    }));
    const missionPresenceArray = Object.keys(missionPresence).map((mc) => ({
      missionId: missionId,
      forceId: mc,
    }));

    await MissionCommandHandler.sync(missionDetails);

    if (differenceWith(missionVehiclesArray, oldMissionVehiclesArray, isEqual).length != 0)
      await MissionVehicleCommandHandler.sync(differenceWith(missionVehiclesArray, oldMissionVehiclesArray, isEqual));
    if (removedVehicles.length != 0)
      await MissionVehicleCommandHandler.delete(removedVehicles.map((rv) => ({ missionId, vehicleId: rv })));

    if (differenceWith(missionEquipmentArray, oldMissionEquipmentArray, isEqual).length != 0)
      await MissionEquipmentCommandHandler.sync(
        differenceWith(missionEquipmentArray, oldMissionEquipmentArray, isEqual)
      );
    if (removedEquipments.length != 0)
      await MissionEquipmentCommandHandler.delete(removedEquipments.map((re) => ({ missionId, equipmentId: re })));

    if (differenceWith(missionContactArray, oldMissionContactArray, isEqual).length != 0)
      await MissionContactCommandHandler.sync(differenceWith(missionContactArray, oldMissionContactArray, isEqual));
    if (removedContacts.length != 0)
      await MissionContactCommandHandler.delete(removedContacts.map((rc) => ({ missionId, contactId: rc })));

    if (differenceWith(missionPresenceArray, oldMissionPresenceArray, isEqual).length != 0)
      await MissionPresenceCommandHandler.sync(differenceWith(missionPresenceArray, oldMissionPresenceArray, isEqual));
    if (removedPresence.length != 0)
      await MissionPresenceCommandHandler.delete(removedPresence.map((mp) => ({ missionId, forceId: mp })));
  }
}
