import { Brackets, Like } from "typeorm";
import { dataSource } from "../../data-source";
import { force } from "../../entity/configuration/force";
import DatabaseActionException from "../../exceptions/databaseActionException";

export default abstract class ForceService {
  /**
   * @description get all forces
   * @returns {Promise<[Array<force>, number]>}
   */
  static async getAll({
    offset = 0,
    count = 25,
    search = "",
    noLimit = false,
    ids = [],
  }: {
    offset?: number;
    count?: number;
    search?: string;
    noLimit?: boolean;
    ids?: Array<string>;
  }): Promise<[Array<force>, number]> {
    let query = dataSource.getRepository(force).createQueryBuilder("force");

    if (search != "") {
      let searchBits = search.split(" ");

      if (searchBits.length < 2) {
        query = query.where(`member.firstname LIKE :searchQuery OR member.lastname LIKE :searchQuery`, {
          searchQuery: `%${searchBits[0]}%`,
        });
      } else {
        searchBits
          .flatMap((v, i) => searchBits.slice(i + 1).map((w) => [v, w]))
          .forEach((term) => {
            query = query
              .orWhere(
                new Brackets((qb) =>
                  qb.where({ firstname: Like(`%${term[0]}%`) }).andWhere({ lastname: Like(`%${term[1]}%`) })
                )
              )
              .orWhere(
                new Brackets((qb) =>
                  qb.where({ firstname: Like(`%${term[1]}%`) }).andWhere({ lastname: Like(`%${term[0]}%`) })
                )
              );
          });
      }
    }

    if (ids.length != 0) {
      query = query.where("force.id IN (:...ids)", { ids: ids });
    }

    if (!noLimit) {
      query = query.offset(offset).limit(count);
    }

    return await query
      .orderBy("force.lastname")
      .addOrderBy("force.firstname")
      .addOrderBy("force.nameaffix")
      .getManyAndCount()
      .then((res) => {
        return res;
      })
      .catch((err) => {
        throw new DatabaseActionException("SELECT", "force", err);
      });
  }

  /**
   * @description get force by id
   * @param {string} id
   * @returns {Promise<force>}
   */
  static async getById(id: string): Promise<force> {
    return dataSource
      .getRepository(force)
      .createQueryBuilder("force")
      .where("force.id = :id", { id: id })
      .getOneOrFail()
      .then((res) => {
        return res;
      })
      .catch((err) => {
        throw new DatabaseActionException("SELECT", "force", err);
      });
  }

  /**
   * @description get available forces
   * @returns {Promise<Array<force>>}
   */
  static async getAvailable(keyDate: Date): Promise<Array<force>> {
    return await dataSource
      .getRepository(force)
      .createQueryBuilder("force")
      .where("force.commissioned <= :keyDate_start", { keyDate_start: keyDate })
      .andWhere(
        new Brackets((qb) =>
          qb
            .where("force.decommissioned IS NULL")
            .orWhere("force.decommissioned >= :keyDate_end", { keyDate_end: keyDate })
        )
      )
      .orderBy("force.lastname")
      .addOrderBy("force.firstname")
      .addOrderBy("force.nameaffix")
      .getMany()
      .then((res) => {
        return res;
      })
      .catch((err) => {
        throw new DatabaseActionException("SELECT", "force", err);
      });
  }
}
