import { Brackets } from "typeorm";
import { dataSource } from "../../data-source";
import { vehicle } from "../../entity/configuration/vehicle";
import DatabaseActionException from "../../exceptions/databaseActionException";

export default abstract class VehicleService {
  /**
   * @description get all vehicles
   * @returns {Promise<[Array<vehicle>, number]>}
   */
  static async getAll({
    offset = 0,
    count = 25,
    search = "",
    noLimit = false,
    ids = [],
  }: {
    offset?: number;
    count?: number;
    search?: string;
    noLimit?: boolean;
    ids?: Array<string>;
  }): Promise<[Array<vehicle>, number]> {
    let query = dataSource.getRepository(vehicle).createQueryBuilder("vehicle");

    if (search != "") {
      search.split(" ").forEach((term, index) => {
        const searchQuery = `%${term}%`;
        const dynamic = "searchQuery" + Math.random().toString(36).substring(2);
        if (index == 0) {
          query = query.where(`vehicle.name LIKE :${dynamic}`, {
            [dynamic]: searchQuery,
          });
        } else {
          query = query.orWhere(`vehicle.name LIKE :${dynamic}`, {
            [dynamic]: searchQuery,
          });
        }
      });
    }

    if (ids.length != 0) {
      query = query.where("vehicle.id IN (:...ids)", { ids: ids });
    }

    if (!noLimit) {
      query = query.offset(offset).limit(count);
    }

    return await query
      .orderBy("vehicle.name")
      .getManyAndCount()
      .then((res) => {
        return res;
      })
      .catch((err) => {
        throw new DatabaseActionException("SELECT", "vehicle", err);
      });
  }

  /**
   * @description get vehicle by id
   * @param {string} id
   * @returns {Promise<vehicle>}
   */
  static async getById(id: string): Promise<vehicle> {
    return dataSource
      .getRepository(vehicle)
      .createQueryBuilder("vehicle")
      .where("vehicle.id = :id", { id: id })
      .getOneOrFail()
      .then((res) => {
        return res;
      })
      .catch((err) => {
        throw new DatabaseActionException("SELECT", "vehicle", err);
      });
  }

  /**
   * @description get available vehicles
   * @returns {Promise<Array<vehicle>>}
   */
  static async getAvailable(keyDate: Date): Promise<Array<vehicle>> {
    return await dataSource
      .getRepository(vehicle)
      .createQueryBuilder("vehicle")
      .where("vehicle.commissioned <= :keyDate_start", { keyDate_start: keyDate })
      .andWhere(
        new Brackets((qb) =>
          qb
            .where("vehicle.decommissioned IS NULL")
            .orWhere("vehicle.decommissioned >= :keyDate_end", { keyDate_end: keyDate })
        )
      )
      .orderBy("vehicle.name")
      .getMany()
      .then((res) => {
        return res;
      })
      .catch((err) => {
        throw new DatabaseActionException("SELECT", "vehicle", err);
      });
  }
}
