import * as Y from "yjs";
import ms from "ms";

export interface missionStoreModel {
  missionId: string;
  doc: Y.Doc;
  editors: Map<string, Editor>;
  editorStates: Map<string, EditorState>;
  timestamp: number;
}

export interface Editor {
  username: string;
  firstname: string;
  lastname: string;
  color: string;
}

export interface EditorState {
  field: string;
  cursor: any;
  range: any;
}

/**
 * @description store credentials to socket to prevent auth data change while connected
 */
export abstract class MissionMap {
  private static store = new Map<string, missionStoreModel>();
  private static timer = new Map<string, NodeJS.Timeout>();
  private static usage = new Map<string, number>();

  public static write(identifier: string, data: missionStoreModel, overwrite: boolean = false): void {
    this.monitorAccess(identifier);

    if (!this.exists(identifier) || overwrite) {
      this.store.set(identifier, data);
    }
  }

  public static updateState(identifier: string, data: Uint8Array): void {
    this.monitorAccess(identifier);

    let mission = this.read(identifier);
    Y.applyUpdate(mission.doc, data);
    this.write(identifier, mission, true);
  }

  public static updateAwareness(
    identifier: string,
    data: { action: "update" | "remove"; socketId: string } & (
      | { store: "editors"; update: Editor }
      | { store: "state"; update: EditorState }
    )
  ): void {
    this.monitorAccess(identifier);

    let mission = this.read(identifier);

    if (data.store == "editors") {
      if (data.action == "update") {
        mission.editors.set(data.socketId, data.update);
      } else {
        mission.editors.delete(data.socketId);
      }
    } else {
      if (data.action == "update") {
        mission.editorStates.set(data.socketId, data.update);
      } else {
        mission.editorStates.delete(data.socketId);
      }
    }

    this.write(identifier, mission, true);
  }

  public static updateTimestamp(identifier: string, data: number): void {
    this.monitorAccess(identifier);

    let mission = this.read(identifier);
    mission.timestamp = data;
    this.write(identifier, mission, true);
  }

  public static read(identifier: string): missionStoreModel {
    this.monitorAccess(identifier);

    return this.store.get(identifier);
  }

  public static exists(identifier: string): boolean {
    this.monitorAccess(identifier);

    return this.store.has(identifier);
  }

  public static delete(identifier: string): void {
    this.abortDelete(identifier);

    const copyIdentifier = identifier;
    const timer = setTimeout(() => {
      this.finalDelete(copyIdentifier);
    }, ms("10s"));

    this.timer.set(identifier, timer);
  }

  public static deleteExpired(): void {
    this.usage.forEach((val, key) => {
      if (Date.now() - val > ms("10s")) {
        this.finalDelete(key);
      }
    });
  }

  private static monitorAccess(identifier: string): void {
    this.usage.set(identifier, Date.now());
    this.abortDelete(identifier);
  }

  private static abortDelete(identifier: string): void {
    const timer = this.timer.get(identifier);
    if (timer) {
      clearTimeout(timer);
      this.timer.delete(identifier);
    }
  }

  private static finalDelete(identifier: string): void {
    const time = this.usage.get(identifier);
    if (Date.now() - time < ms("10s")) return;

    this.store.delete(identifier);
    this.timer.delete(identifier);
    this.usage.delete(identifier);

    console.log(`cleared local doc ${identifier}`);
  }
}
